#!/usr/bin/env python3

#
# Copyright (C) 2020 Cambridge Astronomical Survey Unit
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <https://www.gnu.org/licenses/>.
#




import argparse
import logging
import os
import numpy as _np
import ast

from astropy.io import fits
from astropy.table import Table
from workflow.utils.classes import OBXML

import time

import xml.dom.minidom as minidom
import xml.etree.ElementTree as ET

from collections import defaultdict
from collections import Counter

import traceback
import sys

from concurrent.futures import ProcessPoolExecutor

allowed_modes = ['L', 'O']


def get_group_dict(filename)-> dict:
    """ A function to extract the linking parameters from the CSV file"""
    groups = Table.read(filename, comment='#')
    group_data = {}
	
    for name in groups.colnames:
        try:
            group_data[name] = groups[name].tolist()
            if None in group_data[name]:
                group_data[name] = list(_np.array(group_data[name], dtype=str))
        except SystemExit: 
            logging.error('ERROR: information about {} not found'.format(name))


    return group_data

def check_linking_csv(group_data)-> bool:
    """A function to check the validity of the linking CSV. This includes identifying non-integer priority values 
    and incorrect LINKING_MODE values

    Args:
        group_data (dict): the data of the linking csv file

    Returns:
        bool: whether the csv file has passed the checks
    """
    
    allowed_modes = ['L', 'O']
    expected_elements = 5
    
    for key in group_data:
        if key=='KEY':
            for value in group_data[key]:
                indices = find_key_index(value, group_data)
                if len(indices)>1:
                    logging.warning(f'KEY {value} in linking.csv contains duplicate elements')
                    #return False
                split_elements = value.split(':')
                assert len(split_elements) == expected_elements, f'KEY Structure {value} incorrect'
                
        elif key=='LINKING_MODE':
            for mode in group_data[key]:
                #logging.error('LINKING_MODE must be either L or O')
                assert mode in allowed_modes, 'LINKING_MODE must be either L or O'
                #return False
                        
        elif key=='SEED_VALUE':
            for value in group_data[key]:
                #logging.error('Linking priority must be an integer number')
                assert type(value)==int, 'Linking priority must be an integer number'
                #return False
    


def find_key_index(key, group_data)->list:
    """A function to identify the index needed to extract the linking information for a single key.

    Args:
        key (Str): The key string being searched for
        group_data (dict): The dictionary containing the whole information from the linking csv file

    Returns:
        list: The indicies where the key string has been found. If unique len should equal 1.
    """
    indices = [index for index, value in enumerate(group_data['KEY']) if value == key]
    
    #if len(indices)>1:
        #logging.error('Linking key is not unique')
    if len(indices)==0:
        logging.error('No matching key found in group_data')
        
    return indices

def find_key_info_data(key, group_data)-> dict:
    """A function to generate a dictionary with only the linking information which corresponds to a unique key

    Args:
        key (Str): The key string being searched for
        group_data (dict): The dictionary containing the whole information from the linking csv file

    Returns:
        dict: A dictionary containing the linking information which corresponds to the specific key
    """
    
    key_index = find_key_index(key, group_data)
    ###DO we want an assertion here or just stick with grabbing the first one?
    
    link_data = defaultdict(list)
    try:
        for jkey in key_index:
            for k in group_data:
                link_data[k].append(group_data[k][jkey])
            
    except:
        logging.error('No matching key found')
    
    return link_data
    
def _clear_linking_attributes(obselement)->None:
    """A function to set all of the linking group attributes to an empty string

    Args:
        obselement (_type_): the observation element of the xml file
    """
    
    ###Need to check that I don't have to extract the element before I can set it
    blank_element = ''
    
    old_str_linkgroup = str(obselement.getAttribute('linkedgroup'))
    obselement.setAttribute('linkedgroup', blank_element)
    
    #old_str_linkgroup_prio = str(obselement.getAttribute('linkedgroup_priority'))
    obselement.setAttribute('linkedgroup_priority', blank_element)
    
    #old_str_groupname = str(obselement.getAttribute('obsgroup'))
    obselement.setAttribute('obsgroup', blank_element)
    
    #old_str_group_valid = str(obselement.getAttribute('obsgroup_validity'))
    obselement.setAttribute('obsgroup_validity', blank_element)
    
def _extract_obs_key_elements(obselement):
    obs_key = str(obselement.getAttribute('progtemp')) +':'+ str(obselement.getAttribute('obstemp'))
    
    return obs_key


def _modify_observation_element(obselement, data_dict) -> None:
    
    for key in data_dict:
        value = data_dict.get(key)
        assert not isinstance(value, list), f"Value for key is a list, but should be a single element"
  
    if data_dict['LINKING_MODE'][0]=='L':
            
        old_str_groupname = str(obselement.getAttribute('linkedgroup'))
        new_str_groupname = str(data_dict['GROUP_ID'])
        obselement.setAttribute('linkedgroup', new_str_groupname)
        
        old_str_group_prio = str(obselement.getAttribute('linkedgroup_priority'))
        new_str_group_prio = str(data_dict['SEED_VALUE'])
        obselement.setAttribute('linkedgroup_priority', new_str_group_prio)
            
                
            
    if data_dict['LINKING_MODE'][0]=='O':
        old_str_groupname = str(obselement.getAttribute('obsgroup'))
        new_str_groupname = str(data_dict['GROUP_ID'])
        obselement.setAttribute('obsgroup', new_str_groupname)
        
        old_str_group_valid = str(obselement.getAttribute('obsgroup_validity'))
        new_str_group_valid = str(data_dict['SEED_VALUE'])
        obselement.setAttribute('obsgroup_validity', new_str_group_valid)
    

def find_highest_group_modes(key_list, group_data, O_group_id_counter, L_group_id_counter):
    #
    for index, item in enumerate(key_list):
        mode = ':'.join(item.split(':')[-1])
        KEY = ':'.join(item.split(':')[:-1])
        data = find_key_info_data(KEY, group_data)
        if data:   
            if mode == 'L':
                assert data['LINKING_MODE'][0]=='L', 'Extracted data does not match key mode'
                data_key = data['KEY'][0]
                group_id = data['GROUP_ID'][0]
                if group_id in L_group_id_counter:
                    L_group_id_counter[group_id] +=1
            if mode=='O':
                data_key = data['KEY'][0]
                group_id = data['GROUP_ID'][0]
                if group_id in O_group_id_counter:
                    O_group_id_counter[group_id] +=1
    #Finding the highest one after counters modified
    # Initialize variables to track the highest value and corresponding key
    highest_value_L = float('-inf')  # Start with negative infinity to handle negative values
    highest_group_L = None
    
    highest_value_O = float('-inf')
    highest_group_O = None

    # Iterate through the dictionary and find the highest value (and the last one in case of a tie)
    for group, num in L_group_id_counter.items():
        if num >= highest_value_L:
            highest_value_L = num
            highest_group_L = group
    if highest_value_L == 0:
        highest_group_L = None
    
    for group, num in O_group_id_counter.items():
        if num >= highest_value_O:
            highest_value_O = num
            highest_group_O = group
    
    if highest_value_O == 0:
        highest_group_O = None
                
    logging.info("Highest value LINKING_MODE L->:"+ str(highest_value_L))
    logging.info("GROUP_ID with the highest value (last one in case of a tie):"+str(highest_group_L))
    
    logging.info("Highest value LINKING_MODE O->:"+ str(highest_value_O))
    logging.info("GROUP_ID with the highest value (last one in case of a tie):"+str(highest_group_O))
    return highest_group_L, highest_group_O


def find_entries_by_group_id(data, target_group_id):
    # Initialize a list to store matching entries
    matching_entries = []

    # Iterate through the 'GROUP_ID' list and check for matches
    for key, group_id in zip(data['KEY'], data['GROUP_ID']):
        if group_id == target_group_id:
            matching_entries.append({'KEY': key, 'LINKING_MODE': data['LINKING_MODE'][data['KEY'].index(key)], 'GROUP_ID':data['GROUP_ID'][data['KEY'].index(key)], 'SEED_VALUE': data['SEED_VALUE'][data['KEY'].index(key)]})

    return matching_entries

def check_seed_values(matching_entries):
    # Get the SEED_VALUE of the first entry
    first_seed_value = matching_entries[0]['SEED_VALUE']

    # Check if all matching entries have the same SEED_VALUE
    for entry in matching_entries[1:]:
        if entry['SEED_VALUE'] != first_seed_value:
            raise ValueError("SEED_VALUE mismatch detected!")



###This function may be redundant soon
def find_highest_key(ledge):
    highest_value = float('-inf')  # Start with negative infinity to handle negative values
    keys_with_highest_value = []

    # Iterate through keys and values
    for key, value in ledge.items():
        if value > highest_value:
        # Found a new highest value, update the highest_value and reset keys_with_highest_value
            highest_value = value
            keys_with_highest_value = [key]
        elif value == highest_value:
        # Found another key with the highest value
            keys_with_highest_value.append(key)

    print("Keys with the highest value(s):", keys_with_highest_value)
    print("Highest value:", highest_value)
    return keys_with_highest_value


def open_FITS_file(fits_cat, add_sim):
    """A function to open the catalogue FITS file generated by the WASP with the CNAMES included

    Args:
        fits_cat (Str): The catalogue FITS file output from the WASP
        add_sim (Bool): a flag which states whether simulation data is included in the FITS header

    Returns:
        fits_data (astropy.io.fits.hdu.hdulist.HDUList): the FITS data extracted using astropy.io.fits
    """
    
    with fits.open(fits_cat) as hdu_list:
        
        fits_data = hdu_list[1].data
        
        # Let's check that there are CNAMES in here (a good indication that it
        # has been WASP-validated!)
        assert _np.all(_np.char.startswith(fits_data['CNAME'],'WVE_')), "Not all CNAMEs have been populated in provided FITS catalogue: has this been validated with WASP?"

        
        if add_sim == True:
            try:
                sim_data = hdu_list[2].data
            except Exception as e:
                logging.error('Unable to open sim data from FITS catalogue, please check the correct FITS file is being used')
                logging.error('Exception occurred: '+e)
            
                exc_type, exc_obj, tb = sys.exc_info()
                traceback.print_tb(tb)
                tb_info = traceback.extract_tb(tb)
                filename, line, func, text = tb_info[-1]
                logging.error('An exception occurred on line '+line)
            
        else:
            sim_data = None
        
        return fits_data, sim_data

def generate_output_file_list(xml_files, output_dir):
    """A function to generate the output file list for a series of XML files

    Args:
        xml_files (list): the list of XML files
        output_dir (Str): the directory the output files will be located in

    Returns:
        list: a list of combined XML file names with the corresponding output directory
    """
    return [os.path.join(output_dir, '{}.xml'.format(os.path.basename(xml_file).rsplit('-', 1)[0])) for xml_file in xml_files]

def generate_output_file(xml_file, output_dir):
    """A function to generate the output file path for a specific XML file

    Args:
        xml_files (str): the XML file
        output_dir (Str): the directory the output files will be located in

    Returns:
        str: the combined XML file name with the corresponding output directory
    """
    
    return os.path.join(output_dir, '{}.xml'.format(os.path.basename(xml_file).rsplit('-', 1)[0]))

def outfile_overwrite(output_path, overwrite):
    """
        A function to check if a output file already exists and overwrite it on request
    Args:
        output_path (str): a string stating the output file path
    """
    if os.path.exists(output_path):
            
            if overwrite == True:
                logging.info('Removing previous file: {}'.format(output_path))
                os.remove(output_path)
            else:
                logging.info(
                    'Skipping file because it already exists: {}'.format(
                        output_path))

    else:
        logging.debug('path does not exist')
    return


######functions here to organise the ledger printing
def combine_seed_map(group_data):
    try:
        combined_key_seed_map = {}
        combined_key_group_id_map = {}
        # Iterate through the keys and create the combined keys
        for key, seed_value, linking_mode, group_id in zip(group_data['KEY'], group_data['SEED_VALUE'], group_data['LINKING_MODE'], group_data['GROUP_ID']):
            combined_key = f'{key}:{linking_mode}'
            combined_key_seed_map[combined_key] = seed_value
            combined_key_group_id_map[combined_key] = group_id
            
    except Exception as e:
        logging.error('Unable to recombine group_data for printing formatting')
        logging.error('Exception occurred: '+ e)
            
        exc_type, exc_obj, tb = sys.exc_info()
        traceback.print_tb(tb)
        tb_info = traceback.extract_tb(tb)
        filename, line, func, text = tb_info[-1]
        logging.error('An exception occurred on line'+line)
    
    return combined_key_seed_map, combined_key_group_id_map



def print_ledger_key_files(data, key_seed_map)->None:
    unique_key_files = {}
    combined_key_seed_map, combined_key_group_id_map = combine_seed_map(key_seed_map)
    
    # Iterate through the data
    for filename, key_value_dict in data.items():
        for key, counter in key_value_dict.items():
            if counter > 0:
                if key not in unique_key_files:
                    unique_key_files[key] = []
                unique_key_files[key].append(filename)

    # Print the results
    for key, filenames in unique_key_files.items():
        seed_value = combined_key_seed_map[key]
        group_id = combined_key_group_id_map[key]
        logging.info("Key '{}' with a GROUP_ID of {} and a SEED_VALUE of {} is used in the following files:".format(key,group_id,seed_value))

        for filename in filenames:
            logging.info('\t'+filename)
        logging.info("")

    
    


def process_xml_file(xml_file, fits_data, sim_data, group_alloc, clean_non_allocated, output_dir, overwrite, strict, ledger=None, multiprocess=False):#result_queue):
    """A function to process the XML file and populate it with the appropriate CNAMES generated by the FITS catalogue file

    Args:
        xml_file (Str): the XML filename
        fits_data (astropy.io.fits.hdu.hdulist.HDUList): the FITS data extracted using astropy.io.fits
        sim_data (astropy.io.fits.hdu.hdulist.HDUList): the simulated FITS data extracted using astropy.io.fits
        group_alloc (tuple): Sets any relational linkage between the XML files generated. If empty, no groups are made"
        clean_non_allocated (bool): remove non-allocated targets in the output files
        ob_xml_output (list): a list to append the combined XML file names with the corresponding output directory
    """
    
    #setting start time
    iter_start_time = time.time()
    
    # Create an OB from the XML file
        
    ob_xml = OBXML(xml_file)
    
    ledger[xml_file] = {}
    
    #grabbing the obs linking key info if linking enabled and clear the attributes in both cases
    if group_alloc:
        for element in ob_xml.dom.getElementsByTagName('observation'):
            obs_mode = str(element.getAttribute('obs_mode'))
            _clear_linking_attributes(element)
            obs_key = _extract_obs_key_elements(element) + ':'
            
            logging.debug("obs_key "+obs_key)
        
        #checking the apply dither option
        for element in ob_xml.dom.getElementsByTagName('dithering'):
            ifu_dither = int(element.getAttribute('apply_dither'))
        
    else:
        obs_mode = False
        for element in ob_xml.dom.getElementsByTagName('observation'):
            _clear_linking_attributes(element)
    
    # Populate the OB XML with the FITS data
    
    targ_keys = ob_xml.populate_targets_with_fits_data(fits_data, sim_data=sim_data,
                                            clean_non_allocated=clean_non_allocated, linking=obs_mode)
    

    logging.debug("targ_keys "+str(targ_keys))
    
    # Populate the OB XML with grouping data
    #ob_xml.populate_targets_with_group_data(group_alloc)
    
    
    ## Probably worth trying to abstract this into it's own function
    if group_alloc and isinstance(targ_keys, list) and len(targ_keys)>0:
        try:
            #skipping process if no target keys extracted from the file
            assert len(targ_keys)>=1, 'Cannot process an empty target list'
            
            #Extract keys from group_data dictionary
            #combining LINKING_MODE with the rest of the CSV KEY
            csv_keys = [f"{key}:{linkmode}" for key, linkmode in zip(group_alloc['KEY'], group_alloc['LINKING_MODE'])] 
            # To cut down key back again -> ':'.join(key.split(':')[:-1])
            
            O_group_dict = {group_id: 0 for group_id in group_alloc['GROUP_ID'] if group_alloc['LINKING_MODE'][group_alloc['GROUP_ID'].index(group_id)] == 'O'}
            L_group_dict = {group_id: 0 for group_id in group_alloc['GROUP_ID'] if group_alloc['LINKING_MODE'][group_alloc['GROUP_ID'].index(group_id)] == 'L'}
            
            key_counts = Counter(key for key in csv_keys)
            if strict:
                #Adding assertion that a key can't appear more than once for strict mode
                assert all(count <= 1 for count in key_counts.values()), 'In strict mode: Duplicate linking key is not allowed in this CSV file'
            

            logging.debug("CSV key counts "+str(key_counts))
            
            # Setting the CSV keys inside the ledger dict for this file:
            ledger[xml_file] = {value: 0 for value in set(csv_keys)} #using a set here to make sure the keys aren't repeated in ledger
            
            
            #combining full key together for cross matching from obs and target elements
            targ_keys = [str(obs_key)+str(item)+str(ifu_dither) for index, item in enumerate(targ_keys)]
            full_targ_keys = [item+':'+ m for index, item in enumerate(targ_keys) for m in allowed_modes]
            
            #Adding to the counter if any target keys match those in the CSV file
            #This is structured such that the same key but with 'L' and 'O' increment the same amount
            for key in ledger[xml_file].keys():
                matches = [obj for obj in full_targ_keys if key in obj]
                ledger[xml_file][key] += len(matches)
                
                logging.debug("key_prefix "+ key+  ' '+str(len(matches)))
            
            #setting the number of strict key counter occurences 
            #Should this have a default value of 1?
            #Nstrict = 1
            #if obs_mode == 'mIFU':
            Nstrict = len(targ_keys) #20?
            logging.debug("NStrict "+ str(Nstrict))
            
            #grabbing the list of keys that matched in this file:
            keys_occurred = [key for key, value in ledger[xml_file].items() if value > 0]
            logging.debug('Keys occurred: '+str(keys_occurred))
            # Check if all counters are zero
            all_zero_counters = all(counter == 0 for counter in ledger[xml_file].values())

            if not all_zero_counters:
                #Finding the most grouping in the list
                dominant_L_group, dominant_O_group = find_highest_group_modes(keys_occurred, group_alloc, O_group_dict, L_group_dict)
                #dominant_group = find_highest_group_modes(keys_occurred, group_alloc, O_group_dict, L_group_dict)
                #max_key = max(ledger[xml_file], key=ledger[xml_file].get)
                #max_key = find_highest_key(ledger[xml_file])
                logging.debug("Debug info: Dominant GROUP_IDS "+'L '+str(dominant_L_group)+' and O '+dominant_O_group)
                
                L_entries = find_entries_by_group_id(group_alloc, dominant_L_group)
                O_entries = find_entries_by_group_id(group_alloc, dominant_O_group)
                
                if L_entries:
                    check_seed_values(L_entries)
                if O_entries:
                    check_seed_values(O_entries)
                
                combined_entries = L_entries + O_entries
                '''if strict:
                    #Asserting the number of science bundles that must be in the file
                    for key, value in ledger[xml_file].items():
                        for m in max_key:
                            if m in key:
                                assert value == Nstrict, f"In strict mode: the dominant target key must have the same number of science bundles as mIFU or LIFU. Counter value for key '{key}' doesn't match expected value"
                '''
                
                ##This behaviour takes the dominant key and writes the linking information in for each instance it occurs in the linking CSV file
                #It does not prevent any overwriting unless strict mode is in place
                # for m in max_key:
                #     data_dict =  find_key_info_data(':'.join(m.split(':')[:-1]), group_alloc)
                #     for i in range(0, len(data_dict['KEY'])):
                #         values_dict = {key: values[i] for key, values in data_dict.items()}
                for entry in combined_entries:    
                    logging.debug("Debug info: entry"+ str(entry))
                            
                    for element in ob_xml.dom.getElementsByTagName('observation'):
                        logging.info('MODIFYING file {} WITH VALUES {}'.format(xml_file,entry))
                        _modify_observation_element(element, entry)
            elif all_zero_counters:
                logging.info('No matching keys found in this file')


            
        except Exception as e:
            logging.error('Linking was unable to extract target key info from xml file '+str(xml_file))
            logging.error('Exception occurred: '+e)
            
            exc_type, exc_obj, tb = sys.exc_info()
            traceback.print_tb(tb)
            tb_info = traceback.extract_tb(tb)
            filename, line, func, text = tb_info[-1]
            logging.error('An exception occurred on line '+line)
    
    
    output_path = generate_output_file(xml_file, output_dir)
    
    outfile_overwrite(output_path, overwrite)
    
    ob_xml.write_xml(output_path, replace_tabs=True)
    
    
    #Printing the timings
    iter_end_time = time.time()
    iter_execution_time = iter_end_time - iter_start_time
    logging.info("Execution time: {:.2f} seconds".format(iter_execution_time)) 
    
    if multiprocess:
        return output_path, ledger
    
    return output_path


def run_linear_processing(xml_files, fits_data, sim_data, group_alloc, clean_non_allocated, output_dir, overwrite, strict, ledger):
    results = []
    for xml_file in xml_files:
        process_xml_file(xml_file, fits_data, sim_data, group_alloc, clean_non_allocated, output_dir, overwrite, strict, ledger)
        
    output_paths = [result.get() for result in results]
    
    return output_paths

def run_multiprocessing(xml_files, fits_data, sim_data, group_alloc, clean_non_allocated, output_dir, overwrite, strict, ledger, groups, ncpu=2):


    results = []
    _ledger = []

    jobs = []
    
    with ProcessPoolExecutor(max_workers=ncpu) as executor:
        for xml_file in xml_files:
            job = executor.submit(process_xml_file, xml_file, fits_data, sim_data, group_alloc, clean_non_allocated, output_dir, overwrite, strict, ledger, True)
            jobs.append(job)

    for job in jobs:
        output_path, ledger = job.result()
        results.append(output_path)
        _ledger.append(ledger)


    logging.info('RESULTS: {}'.format(results))
    logging.info('RESULTS: {}'.format(_ledger))

    # Get the output paths from the results
    output_paths = [result for result in results] #[result.get() for result in results]

    return output_paths, _ledger






def fill_xmls_with_fits_info(fits_cat, xml_files, output_dir,
                             group_data=None, clean_non_allocated=True,
                             add_sim=False, overwrite=False):
    """
    Fill a set of XML files with the information from a FITS catalogue.
    
    Parameters
    ----------
    fits_cat : str
        Name of a FITS file containing a FITS catalogue.
    xml_files : list of str
        A list of strings with the filenames of the XML files.
    output_dir : str
        Name of the directory which will contain the output XML files.
    group_data : str, optional
        Location of the CSV file containing grouping data for these XML files.
    clean_non_allocated: bool, optional
        Remove non-allocated targets in the output files.
    add_sim : bool, optional
        Add simulation elements to the targets.
    overwrite : bool, optional
        Overwrite the output XML files.

    Returns
    -------
    output_file_list : list of str
        A list with the output XML files.
    """
        
    # Open the combo FITS catalogue and read the needed data

    with fits.open(fits_cat) as hdu_list:
        
        fits_data = hdu_list[1].data
        
        if add_sim == True:
            sim_data = hdu_list[2].data
        else:
            sim_data = None

    # If it exists, open the group data file and read the data

    group_alloc = {}
    if group_data:
        group_alloc = get_group_dict(group_data)
        


            
    # Let's check that there are CNAMES in here (a good indication that it
    # has been WASP-validated!)
    assert _np.all(_np.char.startswith(fits_data['CNAME'],'WVE_')), "Not all CNAMEs have been populated in provided FITS catalogue: has this been validated with WASP?"

            
    # Create a list for saving the names of the output files
    
    output_file_list = []

    # For each XML file
    logging.info("Updating a total of {} XML files using CNAME'd catalogue {}".format(len(xml_files),fits_cat))
    index = 0

    for xml_file in xml_files:
        iter_start_time = time.time()
        # Create an OB from the XML file
        
        ob_xml = OBXML(xml_file)
        
        # Choose an output filename from the input filename
        
        input_basename = os.path.basename(xml_file)

        output_basename_wo_ext = '-'.join(input_basename.split('-')[:-1])
        
        output_basename = '{}.xml'.format(output_basename_wo_ext)
        
        output_path = os.path.join(output_dir, output_basename)
        
        # Save the name of the output file
        
        output_file_list.append(output_path)

        index += 1
        logging.info('\nPopulating targets for file {} [{}/{}]'.format(
            input_basename,index,len(xml_files)))
        
        # If the output file exists, remove or skip it

        if os.path.exists(output_path):
            
            if overwrite == True:
                logging.info('Removing previous file: {}'.format(output_path))
                os.remove(output_path)
            else:
                logging.info(
                    'Skipping file because it already exists: {}'.format(
                        output_path))
                continue

        # Populate the OB XML with the FITS data

        ob_xml.populate_targets_with_fits_data(
            fits_data, sim_data=sim_data,
            clean_non_allocated=clean_non_allocated)
        
        # Populate the OB XML with grouping data

        logging.info('Setting groups for file {}'.format(output_basename))

        ob_xml.populate_targets_with_group_data(
            group_alloc)

        
        # Write the OB XML to a file
        
        ob_xml.write_xml(output_path, replace_tabs=True)
        
        #checking the timings
        iter_end_time = time.time()
        iter_execution_time = iter_end_time - iter_start_time
        logging.info("Iteration Execution time: {:.2f} seconds".format(iter_execution_time))
    return output_file_list



if __name__ == '__main__':

    parser = argparse.ArgumentParser(description='Fill XMLs with FITS info')

    parser.add_argument('fits_cat',
                        help='a FITS file containing a combo FITS catalogue')

    parser.add_argument('xml_file', nargs='+', help='name of a XML file')

    parser.add_argument('--groups', dest='group_data', default=None,
                        help="""Name of the grouping data file that sets any
                        relational linkage between the XML files generated. If 
                        empty, no groups are made""")
    
    parser.add_argument('--outdir', dest='output_dir', default='output',
                        help="""name of the directory which will contain the
                        output XML files""")

    parser.add_argument('--clean_non_allocated', action='store_true',
                        help='remove non-allocated targets in the output files')
    parser.add_argument('--keep_non_allocated', dest='clean_non_allocated',
                        action='store_false',
                        help='keep non-allocated targets in the output files')
    parser.set_defaults(clean_non_allocated=True)

    parser.add_argument('--add_sim', action='store_true',
                        help='add simulation elements to the targets')

    parser.add_argument('--overwrite', action='store_true',
                        help='overwrite the output files')

    parser.add_argument('--log_level', default='info',
                        choices=['debug', 'info', 'warning', 'error'],
                        help='the level for the logging messages')
    
    parser.add_argument('--strict', action='store_true', help="""Enforce restrictions on the group linking 
                        modifications.""")
    
    parser.add_argument('--multiprocess', action='store_true', help="""Flag to enable multiproccesing to run for stage 9. 
                                                                    This is a work in progress feature""")

    parser.add_argument('--ncpu', type=int, default=1, help="""Set to >1 to enable multiproccesing to run for stage 9""") 

    
    args = parser.parse_args()

    logging.basicConfig(level=getattr(logging, args.log_level.upper()))

    if not os.path.exists(args.output_dir):
        logging.info('Creating the output directory')
        os.mkdir(args.output_dir)

    start_time = time.time()
    
    fits_data, sim_data = open_FITS_file(args.fits_cat, args.add_sim)
    
    
    group_alloc = {}
    if args.group_data:
        group_alloc = get_group_dict(args.group_data)
        check_linking_csv(group_alloc)

    #Old printing for multiprocessing, it doesn't really work so switching to in the loop method
    
    if args.ncpu > 1:
        logging.info("Using multiprocessing with {} CPUs".format(args.ncpu))
        
        ledger = {}
        _, shared_ledger = run_multiprocessing(args.xml_file, fits_data, sim_data, group_alloc, 
                                               args.clean_non_allocated, args.output_dir,
                                               args.overwrite, args.strict, ledger, args.group_data,
                                               ncpu=args.ncpu)

        if args.group_data:
            list_as_string = str(shared_ledger)
            list_of_dicts = ast.literal_eval(list_as_string)

            # Merge the dictionaries into a single dictionary
            merged_dict = {}
            for item in list_of_dicts:
                merged_dict.update(item)

            ledger = merged_dict
                
    else:
        ledger = {}
        run_linear_processing(args.xml_file, fits_data, sim_data, group_alloc, 

                              clean_non_allocated=args.clean_non_allocated, 
                              output_dir=args.output_dir, overwrite=args.overwrite,
                              strict=args.strict, ledger=ledger)


    if args.group_data:
        logging.info('SUMMARY of LINKING:')
        for fname, data in ledger.items():
            logging.info("")
            logging.info("Filename: {}".format(fname))
            for key, value in data.items():
                if value==0:
                    logging.debug("KEY  {} -> Number of links {}".format(key,value))
                else:
                    logging.info("KEY  {} -> Number of links {}".format(key,value))                 

        logging.info("")
        logging.info('KEY USAGE SUMMARY:')
        print_ledger_key_files(ledger, group_alloc)

        

            
    end_time = time.time()
    #calculating and printing the total function run time
    execution_time = end_time - start_time

    logging.info("Execution time: {:.2f} seconds".format(execution_time))

